/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/* Size of the benchmark problem.  The GPU can run larger problems in a
 * reasonable time.
 *
 * For VOLSIZEX, VOLSIZEY, size 1024 is suitable for a few seconds of
 * GPU computation and size 128 is suitable for a few seconds of
 * CPU computation.
 *
 * For ATOMCOUNT, 100000 is suitable for GPU computation and 10000 is
 * suitable for CPU computation.
 */

#include <parboil.h>

#define VOLSIZEX 512
#define VOLSIZEY 512
#define ATOMCOUNT 20000

typedef struct {
  int x;
  int y;
  int z;
} voldim3i;

typedef struct _tag {
  float x;
  float y;
  float z;
  float w;
} float4;

/* The main compute kernel. */
void calc_energy(
    float *energygrid,
    voldim3i grid,
    const float *atoms,
    int numatoms,
    float gridspacing,
    float k,
    struct pb_TimerSet* timers
);

// Solutions
void sol_create_uniform_bin(
    voldim3i grid,
    int num_atoms,
    float gridspacing,
    const float* atom,
    float cutoff
);

void sol_create_neighbor_list(
    float gridspacing,
    float cutoff
);

void sol_calc_energy_with_bins(
    float* energygrid,
    voldim3i grid,
    const float* atoms,
    int num_atoms,
    float gridspacing,
    float cutoff,
    float k
);

