#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cstring>

#include "CPU_kernels.h"
#include "UDTypes.h"

#define max(x,y) ((x<y)?y:x)
#define min(x,y) ((x>y)?y:x)

#define PI 3.14159265359

float kernel_value_CPU(float v){

  float rValue = 0;

  float z = v*v;

  // polynomials taken from http://ccrma.stanford.edu/CCRMA/Courses/422/projects/kbd/kbdwindow.cpp
  float num = (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z*
  (z* 0.210580722890567e-22f  + 0.380715242345326e-19f ) +
   0.479440257548300e-16f) + 0.435125971262668e-13f ) +
   0.300931127112960e-10f) + 0.160224679395361e-7f  ) +
   0.654858370096785e-5f)  + 0.202591084143397e-2f  ) +
   0.463076284721000e0f)   + 0.754337328948189e2f   ) +
   0.830792541809429e4f)   + 0.571661130563785e6f   ) +
   0.216415572361227e8f)   + 0.356644482244025e9f   ) +
   0.144048298227235e10f);

  float den = (z*(z*(z-0.307646912682801e4f)+0.347626332405882e7f)-0.144048298227235e10f);

  rValue = -num/den;

  return rValue;
}

void calculateLUT(float beta, float width, float*& LUT, unsigned int& sizeLUT){
  float v;
  float _width2_4 = (width*width)/4.0;

  if(width > 0){
    // compute size of LUT based on kernel width
    sizeLUT = (unsigned int)(10000*width);

    // allocate memory
    LUT = (float*) malloc (sizeLUT*sizeof(float));

    for(unsigned int k=0; k<sizeLUT; ++k){
      // compute value to evaluate kernel at
      // v in the range 0:(_width/2)^2
      v = (float(k)/float(sizeLUT))*_width2_4;

      // compute kernel value and store
      LUT[k] = kernel_value_CPU(beta*sqrt(1.0-(v/_width2_4)));
    }
  }
}

float kernel_value_LUT(float v, float* LUT, int sizeLUT, float _1_width2_4)
{
  unsigned int k0;
  float v0;

  v *= (float)sizeLUT;
  k0=(unsigned int)(v*_1_width2_4);
  v0 = ((float)k0)/_1_width2_4;
  return  LUT[k0] + ((v-v0)*(LUT[k0+1]-LUT[k0])/_1_width2_4);
}

int gridding_CPU(unsigned int n, parameters params, ReconstructionSample* sample, int* CPUbin, int CPUbin_size, 
		 float* LUT, int sizeLUT, cmplx* gridData[], float* sampleDensity[], int* indeces[]){
  unsigned int NxL, NxH;
  unsigned int NyL, NyH;
  unsigned int NzL, NzH;

  int nx;
  int ny;
  int nz;

  float w;
  unsigned int idx;
  unsigned int idx0;

  unsigned int idxZ;
  unsigned int idxY;

  float Dx2[100];
  float Dy2[100];
  float Dz2[100];
  float *dx2=NULL;
  float *dy2=NULL;
  float *dz2=NULL;

  float dy2dz2;
  float v;

  float kScale[3];
  kScale[0] = float(params.aquisitionMatrixSize[0])/(float(params.reconstructionMatrixSize[0])*float(params.kMax));
  kScale[1] = float(params.aquisitionMatrixSize[1])/(float(params.reconstructionMatrixSize[1])*float(params.kMax));
  kScale[2] = float(params.aquisitionMatrixSize[2])/(float(params.reconstructionMatrixSize[2])*float(params.kMax));

  float maxV = (0.25*params.kernelWidth*params.kernelWidth)-1e-3;

  unsigned int length = (unsigned int) ceil(params.kernelWidth);
  length += (length+1)%2;
  float half_length = float(length)/2.0;

  float beta = PI * sqrt(4*params.kernelWidth*params.kernelWidth/(params.oversample*params.oversample) * (params.oversample-.5)*(params.oversample-.5)-.8);

  float _1_width2_4 = 4.0/(params.kernelWidth*params.kernelWidth);

  unsigned int size_x = params.gridSize[0];
  unsigned int size_y = params.gridSize[1];
  unsigned int size_z = params.gridSize[2];

  float a = kScale[0]*(size_x-1.0)/2.0;
  float b = float(size_x-1)/2.0;

  int gridNumElems = params.gridSize[0]*params.gridSize[1]*params.gridSize[2];

  int pos = 0;
  (*gridData) = (cmplx*) calloc (gridNumElems,sizeof(cmplx));
  (*sampleDensity) = (float*) calloc (gridNumElems,sizeof(float));
  (*indeces) = (int*) malloc (gridNumElems*sizeof(int));
  int* binAlloc = (int*) malloc (gridNumElems*sizeof(int));
  memset(binAlloc, 0xFF, gridNumElems*sizeof(int));

  if (*gridData == NULL || *sampleDensity == NULL || *indeces == NULL){
    printf("unable to allocate temporary CPU space\n");
    exit(1);
  }

  for (int i=0; i < CPUbin_size; i++){
    ReconstructionSample pt = sample[i];

    float kx = (pt.kX*a)+b;
    float ky = (pt.kY*a)+b;
    float kz = (pt.kZ*a)+b;

    NxL = max((kx - half_length), 0);
    NxH = min((kx + half_length), size_x-1.0);

    NyL = max((ky - half_length), 0);
    NyH = min((ky + half_length), size_y-1.0);

    NzL = max((kz - half_length), 0);
    NzH = min((kz + half_length), size_z-1.0);

    if((pt.real != 0.0 || pt.imag != 0.0) && pt.sdc!=0.0)
    {
      for(dz2 = Dz2, nz=NzL; nz<=NzH; ++nz, ++dz2)
      {
        *dz2 = ((kz-nz)*(kz-nz));
      }
      for(dx2=Dx2,nx=NxL; nx<=NxH; ++nx,++dx2)
      {
        *dx2 = ((kx-nx)*(kx-nx));
      }
      for(dy2=Dy2, ny=NyL; ny<=NyH; ++ny,++dy2)
      {
        *dy2 = ((ky-ny)*(ky-ny));
      }

      idxZ = (NzL-1)*size_x*size_y;
      for(dz2=Dz2, nz=NzL; nz<=NzH; ++nz, ++dz2)
      {
        /* linear offset into 3-D matrix to get to zposition */
        idxZ += size_x*size_y;

        idxY = (NyL-1)*size_x;

        /* loop over x indexes, but only if curent distance is close enough (distance will increase by adding x&y distance) */
        if((*dz2)<maxV)
        {
          for(dy2=Dy2, ny=NyL; ny<=NyH; ++ny, ++dy2)
          {
            /* linear offset IN ADDITION to idxZ to get to Y position */
            idxY += size_x;

            dy2dz2=(*dz2)+(*dy2);

            idx0 = idxY + idxZ;

            /* loop over y indexes, but only if curent distance is close enough (distance will increase by adding y distance) */
            if(dy2dz2<maxV)
            {
              for(dx2=Dx2, nx=NxL; nx<=NxH; ++nx, ++dx2)
              {
                /* value to evaluate kernel at */
                v = dy2dz2+(*dx2);

                if(v<maxV)
                {
                  /* kernel weighting value */
                  if (params.useLUT){
                    w = kernel_value_LUT(v, LUT, sizeLUT, _1_width2_4) * pt.sdc;
                  } else {
                    w = kernel_value_CPU(beta*sqrt(1.0-(v*_1_width2_4))) * pt.sdc;
                  }

                  /* linear index of (x,y,z) point */
                  idx = nx + idx0;

                  /* grid data */
                  if(binAlloc[idx] == -1){
                    binAlloc[idx] = pos;
                    (*indeces)[pos] = idx;
                    pos++;
                  }

                  (*gridData)[binAlloc[idx]].real += w*pt.real; //value;
                  (*gridData)[binAlloc[idx]].imag += w*pt.imag;

                  /* estimate sample density */
                  (*sampleDensity)[binAlloc[idx]] += w;
                }
              }
            }
          }
        }
      }
    }
  }

  free(binAlloc);
  return pos;
}

