#include <stdio.h>
#include <cuda_runtime.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "WKFUtils.h"
#include "UDTypes.h"

#include "scanLargeArray.h"
#include "GPU_kernels.cu"
#include "CPU_kernels.h"

#include "cudpp.h"

#define BLOCKSIZE 512
#define PI 3.14159265359

int compare (const void * a, const void * b)
{
  return ( *(int*)a - *(int*)b );
}

void quicksort(int* key, int* value, int l, int r) {
        int i = l-1, j = r, p = l-1, q = r;
        int v = key[r];
        int temp;
        if (r <= l) return;
        for (;;) {
                while (key[++i] < v) ;
                while (v < key[--j]) {
                        if (j == l) break;
                }
                if (i >= j) break;

                temp=key[i]; key[i]=key[j]; key[j]=temp;
                temp=value[i]; value[i]=value[j]; value[j]=temp;

                if (key[i] == v) {
                        p++;
                        temp=key[i]; key[i]=key[p]; key[p]=temp;
                        temp=value[i]; value[i]=value[p]; value[p]=temp;
                }
                if (v == key[j]) {
                        q--;
                        temp=key[q]; key[q]=key[j]; key[j]=temp;
                        temp=value[q]; value[q]=value[j]; value[j]=temp;
                }
        }

        temp=key[i]; key[i]=key[r]; key[r]=temp;
        temp=value[i]; value[i]=value[r]; value[r]=temp;

        j = i-1;
        i = i+1;

        for (int k = l; k < p; k++, j--) {
                temp=key[k]; key[k]=key[j]; key[j]=temp;
                temp=value[k]; value[k]=value[j]; value[j]=temp;
        }

        for (int k = r-1; k > q; k--, i++) {
                temp=key[k]; key[k]=key[i]; key[i]=temp;
                temp=value[k]; value[k]=value[i]; value[i]=temp;
        }

        quicksort(key, value, l, j);
        quicksort(key, value, i, r);
}

void CUDA_interface (unsigned int n, parameters params, ReconstructionSample* sample, float* LUT, int sizeLUT, cmplx* gridData, float* sampleDensity){

  dim3 dims (4,4,4); //size of a gridding block on the GPU

  float kScale[3];
  kScale[0] = float(params.aquisitionMatrixSize[0])/(float(params.reconstructionMatrixSize[0])*float(params.kMax));
  kScale[1] = float(params.aquisitionMatrixSize[1])/(float(params.reconstructionMatrixSize[1])*float(params.kMax));
  kScale[2] = float(params.aquisitionMatrixSize[2])/(float(params.reconstructionMatrixSize[2])*float(params.kMax));

  float beta = PI * sqrt(4*params.kernelWidth*params.kernelWidth/(params.oversample*params.oversample) * (params.oversample-.5)*(params.oversample-.5)-.8);

  int size_x = params.gridSize[0];
  int size_y = params.gridSize[1];
  int size_z = params.gridSize[2];

  int gridNumElems = size_x * size_y * size_z;

  float _1_width2_4 = 4.0 / (params.kernelWidth*params.kernelWidth);

  float maxV = (0.25*params.kernelWidth*params.kernelWidth)-1e-3;

  unsigned int length = (unsigned int) ceil(params.kernelWidth);
  length += (length+1)%2;
  float half_length = float(length)/2.0;

  float a = (kScale[0]*(size_x-1))/2.0;
  float b = (float)(size_x-1)/2.0;
  int size_xy = size_y*size_x;

  float maxVg = float((dims.x-1)*(dims.x-1))+float((dims.y-1)*(dims.y-1))+float((dims.z-1)*(dims.z-1));
  maxVg = sqrt(maxVg);
  maxVg /= 2;
  maxVg = (maxVg+sqrt(maxV)+1.0f)*(maxVg+sqrt(maxV)+1.0f);

  int npad = 0;
  if (n % 64 != 0){
    npad = 64 - (n%64);
  }

  // Host data structures
  int CPUbin_size;
  int* indeces;
  cmplx* gridData_CPU;
  float* sampleDensity_CPU;
  int num=0;
  int cpuStart;
  int* CPUbin;

  // Device data structures
  ReconstructionSample* Sample_d = NULL;
  float* Sample_sd = NULL;
  int* start_d = NULL;
  float2* gridData_d = NULL;
  float* sampleDensity_d = NULL;
  int* idxKey_d = NULL;
  int* idxValue_d = NULL;
  int* numPts_d = NULL;
  samplePtArray sampleArray_d;
  int* cpuStart_d;

  //seperating samples into GPU bins or CPU bins
  wkf_timerhandle timer;
  timer = wkf_timer_create();

  cudaMalloc((void**)&Sample_sd, (n+npad)*sizeof(ReconstructionSample));
  cudaMalloc((void**)&start_d, (gridNumElems+1)*sizeof(int));
  cudaMalloc((void**)&Sample_d, n*sizeof(ReconstructionSample));
  cudaMalloc((void**)&idxKey_d, n*sizeof(int));
  cudaMalloc((void**)&idxValue_d, n*sizeof(int));
  cudaMalloc((void**)&cpuStart_d, sizeof(int));
  cudaMalloc((void**)&numPts_d, gridNumElems*sizeof(int));

  cudaMemset(numPts_d, 0, gridNumElems*sizeof(int));
  cudaMemset(start_d, 0, (gridNumElems+1)*sizeof(int));
  cudaMemcpyToSymbol(maxVg_c, &maxVg, sizeof(float), 0);
  cudaMemcpyToSymbol(maxV_c, &maxV, sizeof(float), 0);
  cudaMemcpyToSymbol(half_length_c, &half_length, sizeof(float), 0);
  cudaMemcpyToSymbol(gridSize_c, params.gridSize, 3*sizeof(int), 0);
  cudaMemcpyToSymbol(size_xy_c, &size_xy, sizeof(int), 0);
  cudaMemcpyToSymbol(_1_width2_4_c, &_1_width2_4, sizeof(float), 0);
  cudaMemcpy(Sample_d, sample, n*sizeof(ReconstructionSample), cudaMemcpyHostToDevice);

  sampleArray_d.real = Sample_sd;
  sampleArray_d.imag = Sample_sd+(n+npad);
  sampleArray_d.kx = Sample_sd+2*(n+npad);
  sampleArray_d.ky = Sample_sd+3*(n+npad);
  sampleArray_d.kz = Sample_sd+4*(n+npad);
  sampleArray_d.sdc = Sample_sd+5*(n+npad);

  wkf_timer_start(timer);

  dim3 block1 (512);
  dim3 grid1 ((n+511)/512);

  binning_kernel<<<grid1, block1>>>(n, a, b, Sample_d, idxKey_d, idxValue_d, numPts_d, params.binsize, gridNumElems);
  
  if(cudaGetLastError())
	printf("Unable to launch binning kernel\n");

  if (params.sync){
    cudaThreadSynchronize();
  }

  CUDPPConfiguration config;
  config.datatype = CUDPP_UINT;
  config.algorithm = CUDPP_SORT_RADIX;
  config.options = CUDPP_OPTION_KEY_VALUE_PAIRS;

  CUDPPHandle sortplan = 0;
  CUDPPResult result = cudppPlan(&sortplan, config, n, 1, 0);

  double precision = log(double(gridNumElems+1));
  precision /= log(2.0);
  precision = ceil(precision);

  cudppSort(sortplan, idxKey_d, idxValue_d, int(precision), n);
  
  result = cudppDestroyPlan(sortplan);

  if (params.sync){
    cudaThreadSynchronize();
  }

  reorder_kernel<<<grid1,block1>>>(n, idxKey_d, idxValue_d, Sample_d, sampleArray_d, start_d, a, b, gridNumElems, cpuStart_d);
  
  if(cudaGetLastError())
  	printf("Unable to launch reorder kernel\n");

  if (params.sync){
    cudaThreadSynchronize();
  }

  cudaMemcpy(&cpuStart, cpuStart_d, sizeof(int), cudaMemcpyDeviceToHost);

  CPUbin_size = n-cpuStart;
  if (CPUbin_size > 0){
    CPUbin = (int*) malloc (CPUbin_size*sizeof(int));
    cudaMemcpy(CPUbin, idxValue_d, CPUbin_size*sizeof(int), cudaMemcpyDeviceToHost);
  }

  cudaFree(numPts_d);
  cudaFree(cpuStart_d);
  cudaFree(idxValue_d);
  cudaFree(idxKey_d);
  cudaFree(Sample_d);

  scanLargeArray(gridNumElems+1, start_d);

  if (params.sync){
    cudaThreadSynchronize();
  }

  wkf_timer_stop(timer);
  printf("  Total binning time is %f\n", wkf_timer_time(timer));
  
  wkf_timer_start(timer);

  cudaMalloc((void**)&gridData_d, gridNumElems*sizeof(float2));
  cudaMalloc((void**)&sampleDensity_d, gridNumElems*sizeof(float));

  cudaMemset(gridData_d, 0, gridNumElems*sizeof(float2));
  cudaMemset(sampleDensity_d, 0, gridNumElems*sizeof(float));

  dim3 block2 (dims.x,dims.y,dims.z);
  dim3 grid2 (size_x/dims.x, (size_y*size_z)/(dims.y*dims.z));

  gridding_GPU<<<grid2, block2>>>(sampleArray_d, start_d, gridData_d, sampleDensity_d, beta);
  
  if(cudaGetLastError())
	printf("Unable to launch gridding kernel\n");

  if (params.sync){
    cudaThreadSynchronize();
  }

  wkf_timer_stop(timer);
  printf("  Time taken for GPU computing is %f\n", wkf_timer_time(timer));
  wkf_timer_start(timer);

  printf("  Size of CPU workload is %d samples\n", CPUbin_size);

  qsort(CPUbin, CPUbin_size, sizeof(int), compare);
  num = gridding_CPU(n, params, sample, CPUbin, CPUbin_size, LUT, sizeLUT, &gridData_CPU, &sampleDensity_CPU, &indeces);

  wkf_timer_stop(timer);
  printf("  Time taken for CPU computing is %f\n", wkf_timer_time(timer));

  cudaMemcpy(sampleDensity, sampleDensity_d, gridNumElems*sizeof(float),cudaMemcpyDeviceToHost);
  cudaMemcpy(gridData, gridData_d, gridNumElems*sizeof(float2),cudaMemcpyDeviceToHost);

  for (int i=0; i<num; i++){
    gridData[indeces[i]].real += gridData_CPU[i].real;
    gridData[indeces[i]].imag += gridData_CPU[i].imag;
    sampleDensity[indeces[i]] += sampleDensity_CPU[i];
  }

  if (gridData_CPU != NULL){
    free(indeces);
    free(gridData_CPU);
    free(sampleDensity_CPU);
  }

  cudaFree(CPUbin);
  cudaFree(gridData_d);
  cudaFree(sampleDensity_d);
  cudaFree(start_d);
  cudaFree(Sample_sd);

  wkf_timer_destroy(timer);

  return;
}
