#include <stdio.h>
#include <cuda_runtime.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "UDTypes.h"

#define TILE 64
#define LOG_TILE 6

__constant__ float maxV_c;
__constant__ float maxVg_c;
__constant__ float half_length_c;
__constant__ int gridSize_c[3];
__constant__ int size_xy_c;
__constant__ float _1_width2_4_c;

/*__global__ void set_binStart(int gridNumElems, int binsize, int* start_g){
  int idx = blockIdx.x*512 + threadIdx.x;

  if(idx < (gridNumElems+1)){
    start_g[idx] = binsize*idx;
  }
}*/

__global__ void binning_kernel (int n, float a, float b, ReconstructionSample* Sample_g, int* idxKey_g,
                                int* idxValue_g, int* numPts_g, int binsize, int gridNumElems){
  int key;
  int idx = blockIdx.x*blockDim.x+threadIdx.x;
  ReconstructionSample pt;
  int binIdx;
  int cap;

  if (idx < n){
    pt = Sample_g[idx];

    pt.kX = (pt.kX*a)+b;
    pt.kY = (pt.kY*a)+b;
    pt.kZ = (pt.kZ*a)+b;

    binIdx = int(pt.kZ)*size_xy_c + int(pt.kY)*gridSize_c[0] + int(pt.kX);
    cap = atomicAdd(&(numPts_g[binIdx]), 1);

    if(cap < binsize){
      key = binIdx;
    } else {
      key = gridNumElems;
    }

    idxKey_g[idx] = key;
    idxValue_g[idx] = idx;
  }
}

__global__ void reorder_kernel(int n, int* key_g, int* value_g, ReconstructionSample* Samples_g, samplePtArray SampleArray_g, int* start_g,
                               float a, float b, int gridNumElems, int* cpuStart_g){
  int index = blockIdx.x*blockDim.x + threadIdx.x;
  int gridIdx = key_g[index];
  ReconstructionSample pt;

  if (index < n && gridIdx != gridNumElems){
    int old_index = value_g[index];
    pt = Samples_g[old_index];
    SampleArray_g.real[index] = pt.real;
    SampleArray_g.imag[index] = pt.imag;
    SampleArray_g.kx[index] = (pt.kX*a)+b;
    SampleArray_g.ky[index] = (pt.kY*a)+b;
    SampleArray_g.kz[index] = (pt.kZ*a)+b;
    SampleArray_g.sdc[index] = pt.sdc;
  }

  if (index == 0){
    if(gridIdx == gridNumElems){
      cpuStart_g[0] = index;
    }
  }

  if (index < n-1){
    if (gridIdx != key_g[index+1]){
      start_g[gridIdx] = index+1;
      if(key_g[index+1] == gridNumElems){
        cpuStart_g[0] = index+1;
      }
    }
  } else if (index == n-1){
    start_g[gridIdx] = n;
    if(gridIdx != gridNumElems){
      cpuStart_g[0] = n;
    }
  }

}

__device__ float kernel_value(float v){

  float rValue = 0;

  float z = v*v;

  // polynomials taken from http://ccrma.stanford.edu/CCRMA/Courses/422/projects/kbd/kbdwindow.cpp
  float num = (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z*
                (z* 0.210580722890567e-22f  + 0.380715242345326e-19f ) +
                 0.479440257548300e-16f) + 0.435125971262668e-13f ) +
                 0.300931127112960e-10f) + 0.160224679395361e-7f  ) +
                 0.654858370096785e-5f)  + 0.202591084143397e-2f  ) +
                 0.463076284721000e0f)   + 0.754337328948189e2f   ) +
                 0.830792541809429e4f)   + 0.571661130563785e6f   ) +
                 0.216415572361227e8f)   + 0.356644482244025e9f   ) +
                 0.144048298227235e10f);

  float den = (z*(z*(z-0.307646912682801e4f)+0.347626332405882e7f)-0.144048298227235e10f);

  rValue = __fdividef(-num,den);

  return rValue;
}

__global__ void gridding_GPU (samplePtArray sampleArray_g, int* start_g, float2* gridData_g, float* sampleDensity_g, float beta){
  __shared__ float real_s[TILE];
  __shared__ float imag_s[TILE];
  __shared__ float kx_s[TILE];
  __shared__ float ky_s[TILE];
  __shared__ float kz_s[TILE];
  __shared__ float sdc_s[TILE];

  __shared__ int x0;
  __shared__ int y0;
  __shared__ int z0;
  __shared__ int xL;
  __shared__ int xH;
  __shared__ int yL;
  __shared__ int yH;
  __shared__ int zL;
  __shared__ int zH;

  int flatIdx = threadIdx.z*blockDim.y*blockDim.x+threadIdx.y*blockDim.x+threadIdx.x;

  // figure out starting point of the tile
  if (flatIdx == 0) {
    z0 = blockDim.z*(blockIdx.y/(gridSize_c[1]/blockDim.y));
    y0 = blockDim.y*(blockIdx.y%(gridSize_c[1]/blockDim.y));
    x0 = blockIdx.x*blockDim.x;

    xL = x0-(half_length_c+1);
    if (xL < 0){ xL = 0; }
    xH = x0+blockDim.x+half_length_c+1;
    if (xH >= gridSize_c[0]) { xH = gridSize_c[0]-1; }

    yL = y0-(half_length_c+1);
    if (yL < 0){ yL = 0; }
    yH = y0+blockDim.y+half_length_c+1;
    if (yH >= gridSize_c[1]) { yH = gridSize_c[1]-1; }

    zL = z0-(half_length_c+1);
    if (zL < 0){ zL = 0; }
    zH = z0+blockDim.z+half_length_c+1;
    if (zH >= gridSize_c[2]) { zH = gridSize_c[2]-1; }
  }
  __syncthreads();

  float2 pt;
  pt.x = 0.0;
  pt.y = 0.0;
  float density = 0.0;

  for (int z = zL; z <= zH; z++){
      for (int y = yL; y <= yH; y++){

          int start = start_g[z*size_xy_c+ y*gridSize_c[0] + xL];
          int end = start_g[z*size_xy_c+ y*gridSize_c[0] + xH + 1];
          int delta = end-start;
          for (int x = 0; x < ((delta+TILE-1)>>LOG_TILE); x++){
            int fIdx = flatIdx+(x<<LOG_TILE);
            if(fIdx < delta){
              real_s[flatIdx] = sampleArray_g.real[start+fIdx];
              imag_s[flatIdx] = sampleArray_g.imag[start+fIdx];
              kx_s[flatIdx] = sampleArray_g.kx[start+fIdx];
              ky_s[flatIdx] = sampleArray_g.ky[start+fIdx];
              kz_s[flatIdx] = sampleArray_g.kz[start+fIdx];
              sdc_s[flatIdx] = sampleArray_g.sdc[start+fIdx];
            }
            __syncthreads();

            int jH = ((delta-(x<<LOG_TILE)) > TILE) ? TILE : (delta-(x<<LOG_TILE));
            for (int j=0; j< jH; j++){
	      if(real_s[j] != 0.0 || imag_s[j] != 0.0){
                float w;
		float v = 0.0;
                v += (kx_s[j]-(x0+threadIdx.x))*(kx_s[j]-(x0+threadIdx.x));
                v += (ky_s[j]-(y0+threadIdx.y))*(ky_s[j]-(y0+threadIdx.y));
                v += (kz_s[j]-(z0+threadIdx.z))*(kz_s[j]-(z0+threadIdx.z));

                if(v<maxV_c){
                  w = kernel_value(beta*sqrtf(1.0-(v*_1_width2_4_c))) *sdc_s[j];
                  pt.x += w*real_s[j];
                  pt.y += w*imag_s[j];
                  density += w;
                }
	      }
            }
            __syncthreads();
          }
      }
  }

  int idx = (z0 + threadIdx.z)*size_xy_c + (y0 + threadIdx.y)*gridSize_c[0] + x0 + threadIdx.x;
  gridData_g[idx] = pt;
  sampleDensity_g[idx] = density;
}

